# REALSpaceCSVHelper.py - Data handling routines developed by Optibrium for integration
# with StarDrop software
# Use of this code is subject to the terms of the StarDrop License Agreement and
# use, copying or redistribution of this file for use by other persons or
# organisations is prohibited
#
# Copyright(C) Optibrium Ltd 2017
# stardrop-support@optibrium.com

"""
Module provides some functions to assist in parsing CSV type data into a
StarDrop data set.
"""
import app
import string
import io
import csv

class ColumnInfo:
    """
    A small class to represent a column of information to be loaded into StarDrop.
    """

    def __init__(self, name, index):
        """
        Create a ColumnInfo object with a name and column index.

        @param: name: The name of the column to appear in StarDrop.
        @param: index: The column index of the columns data in the CVS data.
        """
        self.name = name
        self.columnIndex = index
        self.errorIndex = -1

    def __repr__(self):
        return "Name: %s, index %i, error %i" % (self.name, self.columnIndex, self.errorIndex)


def checkType(values, name, config):
    """
    Deduce the desired type from a column of data. Assumes data is continuous by
    default.

    @param: values: A list of strings representing the data in the column.
    @param: name: The name of the column.
    @param: config: A configuration module that can be used to control the process.
    @returns: A StarDrop data type (cont, mol, or string)
    """
    # check for molecules/structures
    rv = app.cont

    if 'name' in name:
        rv = app.string
    elif 'smiles' in name:
        rv = app.mol
    else:

        if name.upper().find("STRUCT") != -1 or name.upper().find("SMILES") != -1:
            rv = app.mol
            return rv
        # check for numerical values - check first TYPE_CHECK_DEPTH, and look for
        # at least TYPE_CHECK_FRACTION to be numbers
        count = 0.0
        numbers = 0.0
        index = 0
        while count < config.TYPE_CHECK_DEPTH:
            if index < len(values):
                # check for a null - count as valid number
                if values[index] is None or values[index] == '':
                    pass
                else:
                    count += 1.0
                    try:
                        v = float(values[index])
                        numbers += 1.0
                    except:
                        if values[
                                index].lower() == 'invalid': # count this entry as
                            # representing a number column with invalid entries
                            numbers += 1.0
                        elif len(values[index]) > 0 and values[index][0] in (
                                '<', '>', '='):  # count a EQ, GT or LT operator as
                            # being a number
                            numbers += 1.0
                        pass
                index += 1
            else:
                break
        if count > 0:
            if numbers / count >= config.TYPE_CHECK_FRACTION:
                rv = app.cont
            else:
                rv = app.string
    return rv


def extractResult(result):
    """
    Extract a value and a qualifier from an input string.

    @param: result: The inpiut string.
    @returns: A tuple of qualifier and value (as a float)
    """
    # if the data is empty or invalid at this point an exception will be thrown
    # thrown and caught in loadData
    qual = '\x00'
    value = 0.0
    main = result.rstrip(string.letters)
    if main[0] == '<':
        qual = '\x01'
        main = main.lstrip('<')
        main = main.lstrip()
    elif main[0] == '>':
        qual = '\x02'
        main = main.lstrip('>')
        main = main.lstrip()
    elif main[0] == '=':
        main = main.lstrip('=')
        main = main.lstrip()
    value = float(main)
    return qual, value


def extractValueFromList(results):
    """
    Get a data entry from a list of values by calculating a mean.

    @param: results: The values as strings.
    @returns: A numeric data entry containing the average value, or an invalid entry.
    """
    count = 0
    total = 0.0
    for r in results:
        try:
            value = float(r.lstrip().rstrip())
            total += value
            count += 1
        except:
            pass
    if count > 0:
        return app.Dataset.cont(total / count, 0.0)
    else:
        return app.Dataset.createInvalidEntry(app.cont)


def createContEntry(entry, errorEntry='0.0'):
    """
    Create a numeric entry from strings for the value and the error.

    @param: value: The value as a string , which may include a qualifier.
    @param: errorEntry: The error value as a string.
    @returns: A numeric data entry for a StarDrop data set.
    """
    value = 0.0
    error = 0.0
    qualifier = '\x00'
    try:
        error = float(errorEntry)
        value = float(entry)
    except:
        results = entry.split(',')
        if len(results) > 1:
            return extractValueFromList(results)
        qualifier, value = extractResult(results[0].lstrip().rstrip())
        pass
    return (12 + 2 * 8 + 1, (True, [], []), (qualifier, value, error))


def createDataSet(output, config):
    """
    Take the content of a csv file or csv-like output from another source,
    and parse into a StarDrop data set.

    @param: output: The csv data as a string.
    @param: config: A configuration module that can be used to control the process.
    @returns: A python representation of a StarDrop data set.
    """

    with open(output, 'rb') as csvfile:
        csv_reader = csv.reader(csvfile, delimiter=',', quotechar='"')
        first = True
        headers = list()
        table = list()

        for row in csv_reader:
            if first:
                first = False
                i = 0
                for header in row:
                    table.append(list())
                    headers.append(ColumnInfo(header, i))
                    i += 1
            else:
                if len(row) > len(headers):
                    raise Exception('Incomplete data')
                elif len(row) < len(headers):
                    while len(row) < len(headers):
                        row.append('')

                i = 0
                for item in row:
                    table[i].append(item)
                    i += 1

        try:
            import ColumnFormatter
            headers = ColumnFormatter.formatColumns(headers)
        except:
            pass

        # now build the columns from the data in the table
        ds = app.Dataset()
        ds.cols = []
        for header in headers:
            if header.name.rstrip('"').lstrip('"').decode('latin-1') == '#result-smiles':
                header_corrected = 'Result structure'
            elif header.name.rstrip('"').lstrip('"').decode('latin-1') == 'result-rank':
                header_corrected = 'Result rank'
            elif header.name.rstrip('"').lstrip('"').decode('latin-1') == 'similarity':
                header_corrected = 'Similarity'
            elif header.name.rstrip('"').lstrip('"').decode('latin-1') == 'result-name':
                header_corrected = 'REAL space name'
            elif header.name.rstrip('"').lstrip('"').decode('latin-1') == 'query-smiles':
                header_corrected = 'Query SMILES'
            elif header.name.rstrip('"').lstrip('"').decode('latin-1') == 'similarity-descriptor-smiles':
                header_corrected = 'similarity-descriptor-smiles'
            elif header.name.rstrip('"').lstrip('"').decode('latin-1') == 'similarity-descriptor':
                header_corrected = 'similarity-descriptor'
            elif header.name.rstrip('"').lstrip('"').decode('latin-1') == 'source':
                header_corrected = 'Source'
            elif header.name.rstrip('"').lstrip('"').decode('latin-1') == 'reaction-name':
                header_corrected = 'Reaction name'
            elif header.name.rstrip('"').lstrip('"').decode('latin-1') == 'reagent1-name':
                header_corrected = 'Reagent 1 name'
            elif header.name.rstrip('"').lstrip('"').decode('latin-1') == 'reagent1-smiles':
                header_corrected = 'Reagent 1 SMILES'
            elif header.name.rstrip('"').lstrip('"').decode('latin-1') == 'reagent2-name':
                header_corrected = 'Reagent 2 name'
            elif header.name.rstrip('"').lstrip('"').decode('latin-1') == 'reagent2-smiles':
                header_corrected = 'Reagent 2 SMILES'
            elif header.name.rstrip('"').lstrip('"').decode('latin-1') == 'reagent3-name':
                header_corrected = 'Reagent 3 name'
            elif header.name.rstrip('"').lstrip('"').decode('latin-1') == 'reagent3-smiles':
                header_corrected = 'Reagent 3 SMILES'
            elif header.name.rstrip('"').lstrip('"').decode('latin-1') == 'reagent4-name':
                header_corrected = 'Reagent 4 name'
            elif header.name.rstrip('"').lstrip('"').decode('latin-1') == 'reagent4-smiles':
                header_corrected = 'Reagent 4 SMILES'
            elif header.name.rstrip('"').lstrip('"').decode('latin-1') == 'reagent5-name':
                header_corrected = 'Reagent 5 name'
            elif header.name.rstrip('"').lstrip('"').decode('latin-1') == 'reagent5-smiles':
                header_corrected = 'Reagent 5 SMILES'
            col = []
            if header.name.rstrip('"').lstrip('"').decode('latin-1') in ('similarity-descriptor', 'similarity-descriptor-smiles'):
                ds.columns.append(
                    app.Dataset.meta(header.name.rstrip('"').lstrip('"').decode('latin-1'), app.string, ''))
            else:
                if len(table[:-1]) > 0:
                    type = app.cont
                    if header.errorIndex == -1:
                        type = checkType(table[header.columnIndex], header.name, config)
                    details = ''
                    try:
                        if type == app.cont:
                            details = ((0.0, 0.0), (0, 0))
                        try:
                            ds.columns.append(
                                app.Dataset.meta(header_corrected.decode('latin-1'), type, details))
                        except:
                            ds.columns.append(
                                app.Dataset.meta(header.name.rstrip('"').lstrip('"').decode('latin-1'), type, details))

                        if header.errorIndex == -1:
                            for entry in table[header.columnIndex]:

                                if entry == None:
                                    entry = ''
                                if type == app.mol:
                                    entry = entry.replace('[R1*]', '*').replace('[R2*]', '*').replace('[R3*]', '*').replace('[R4*]', '*').replace('[R5*]', '*')
                                    try:
                                        col.append(app.Dataset.molecule(entry))
                                    except Exception as e:
                                        print(e)
                                elif type == app.string:

                                    if header_corrected == 'REAL space name':
                                        if "__" in entry:
                                            None
                                        elif entry == "":
                                            None
                                        else:
                                            entry = '<a href="https://www.enaminestore.com/catalog/{0}">{0}</a>'.format(
                                                entry)

                                    col.append(app.Dataset.string(entry))
                                else:
                                    try:
                                        col.append(createContEntry(entry))
                                    except Exception, e:
                                        col.append(app.Dataset.createInvalidEntry(app.cont))
                                        if str(e).find("Entry contains a list") != -1:
                                            raise
                        else:
                            for value, error in zip(table[header.columnIndex], table[header.errorIndex]):
                                try:
                                    col.append(createContEntry(value, error))
                                except:
                                    col.append(app.Dataset.createInvalidEntry(app.cont))
                    except Exception, e:
                        # add as a string column instead
                        print(e)
                        ds.columns.pop()
                        details = ''
                        col = []
                        try:
                            ds.columns.append(
                                app.Dataset.meta(header_corrected.decode('latin-1'), app.string, details))
                        except:
                            ds.columns.append(
                                app.Dataset.meta(header.name.rstrip('"').lstrip('"').decode('latin-1'), app.string,
                                                 details))
                        for entry in table[header.columnIndex]:
                            if entry == None:
                                entry = ''
                            col.append(app.Dataset.string(entry))
                else:
                    try:
                        ds.columns.append(
                            app.Dataset.meta(header_corrected.decode('latin-1'), app.string, ''))
                    except:
                        ds.columns.append(
                            app.Dataset.meta(header.name.rstrip('"').lstrip('"').decode('latin-1'), app.string, ''))
            ds.cols.append(col)

        list_of_cols_to_del = []

        # identify any empty columns, put them into a list
        for i in range(len(headers)):
            valid_entry = False
            if i < 4:
                valid_entry = True
            else:
                for field in ds.cols[i]:
                    if field[2] == '\x00':
                        None
                    elif field[2][0] == "":
                        None
                    else:
                        valid_entry = True

            if valid_entry:
                None
            else:
                list_of_cols_to_del.append(i)

        # reverse the list of empty columns to delete them without affecting indexing
        reverse_list = list_of_cols_to_del[::-1]

        # delete the empty columns
        if reverse_list:
            for col_num in reverse_list:
                ds.delColumn(col_num)
        else:
            None

        return ds

